/*
 * FindBugs - Find bugs in Java programs
 *
 * (C) Copyright 2005 Diomidis Spinellis <dds@aueb.gr>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: ApiVerify.java,v 1.11 2005/08/09 10:53:12 dds Exp $
 */

package edu.umd.cs.findbugs.detect;

import java.lang.reflect.Constructor;
import java.lang.reflect.Array;

import java.util.HashMap;
import java.util.HashSet;
import java.util.EnumMap;

import org.apache.bcel.classfile.Code;
import org.apache.bcel.classfile.JavaClass;
import org.apache.bcel.classfile.Method;

import edu.umd.cs.findbugs.BugInstance;
import edu.umd.cs.findbugs.BugReporter;
import edu.umd.cs.findbugs.BytecodeScanningDetector;
import edu.umd.cs.findbugs.OpcodeStack;
import edu.umd.cs.findbugs.MethodAnnotation;
import edu.umd.cs.findbugs.StatelessDetector;
import edu.umd.cs.findbugs.Detector;
import edu.umd.cs.findbugs.ba.ClassContext;
import edu.umd.cs.findbugs.visitclass.Constants2;
import edu.umd.cs.findbugs.apiv.VerifyResult;

/**
 * Generic API invocation verifier.
 * This verifier class will try to locate for every method invocation
 * a method with the same name in a class named with the name of the
 * original class with V appended to it (for example MathV).
 * The verificationclass is searched in the same package as the original class,
 * and in package named edu.umd.cs.findbugs.apiv followed by the package
 * of the original class (for example edu.umd.cs.findbugs.apiv.java.lang).
 * It will then invoke the corresponding method passing as the
 * first argument a Boolean array indicating which of the arguments are
 * compile-time constants, and as the second argument the invocation's object.
 * Any following arguments are those specified in the original invocation.
 * The method's implementation can then verify the argument values
 * according to the API's specification, and return one of the
 * result values defined in VerifyResult.
 *
 * @see edu.umd.cs.findbugs.apiv.VerifyResult
 * @author  Diomidis Spinellis
 * @version $Revision: 1.11 $
 */

public class ApiVerify
extends BytecodeScanningDetector implements Constants2, StatelessDetector {
    private static final boolean DEBUG
	= Boolean.getBoolean("apiverify.debug");
    private static final boolean showInvocationParameters
	= Boolean.getBoolean("apiverify.showInvocationParameters");
    private static final boolean showPerformanceMetrics
	= Boolean.getBoolean("apiverify.showPerformanceMetrics");
    private static final boolean showAllInvocations
	= Boolean.getBoolean("apiverify.showAllInvocations");

    /** Performance metrics collected for every verification method invocation. */
    private static HashMap<String, EnumMap<VerifyResult.Value, Integer>> performanceMetrics =
        new HashMap<String, EnumMap<VerifyResult.Value, Integer>>();
    private static HashSet<String> uniqueInvocations = new HashSet<String>();
    private static int totalInvocations;

    BugReporter bugReporter;

    public ApiVerify(BugReporter bugReporter) {
        this.bugReporter = bugReporter;
    }

    public Object clone() throws CloneNotSupportedException {
        return super.clone();
    }

    public void visit(JavaClass obj) {
    }

    public void visit(Method obj) {
    }

    OpcodeStack stack = new OpcodeStack();
    public void visit(Code obj) {
	stack.resetForMethodEntry(this);
        super.visit(obj);
    }

    /** Return the number of arguments specified in the
     * original method descriptor.
     */
    private int getNArgs(String methodDesc) {
	int args = 0;
        // Loop over characters in methodName; avoid initial (.
        loop: for (int c = 1; c < methodDesc.length(); c++) {
            switch (methodDesc.charAt(c)) {
	    case ')':
	    	break loop;
            case 'B': case 'C': case 'D': case 'F':
            case 'I': case 'J': case 'S': case 'Z':
	    	args++;
		break;
            case '[':
		break;
            case 'L':
                c += methodDesc.indexOf(';', c) - (c + 1);
		args++;
                break;
            }
        }
	if (DEBUG)
	    System.out.println("Method [" + methodDesc + "] takes " + args + " arguments");
	return args;
    }

    /** Finds a verifier method of the specified class with the given
     * name and original method descriptor.
     *
     * @param clazz the name of the verifier method's class
     * @param methodName the method name
     * @param methodDesc the original method descriptor
     * @param argType an array containing the classes of the argument
     * types.  Upon return this array, starting from the item pointed
     * by the offset, will be filled with the appropriate arguments.
     * Existing entries from the offset onwards will be overwritten.
     * @param offset the starting point for updating the argType array
     * (inclusive).
     * @param decl if true, the method might not be publicly accessible
     *
     * @return the verifier method found
     *
     * @throws ClassNotFoundException when clazz cannot be found
     * @throws NoSuchMethodException when the method cannot be found
     */
    private java.lang.reflect.Method getVMethod(Class clazz,
        String methodName, String methodDesc, Class[] argType,
        int offset, boolean decl)
        throws ClassNotFoundException, NoSuchMethodException {
        int i = offset;
        int c = 1; // char counter in methodName, avoid initial (
        int dim = 0; // number of dimensions in array parameter descriptor
        while (c < methodDesc.length()) {
            char ch = methodDesc.charAt(c);
            Class arg = null;
            if (ch == ')') break;
            switch (ch) {
            case 'B': arg = Boolean.TYPE; break;
            case 'C': arg = Character.TYPE; break;
            case 'D': arg = Double.TYPE; break;
            case 'F': arg = Float.TYPE; break;
            case 'I': arg = Integer.TYPE; break;
            case 'J': arg = Long.TYPE; break;
            case 'S': arg = Short.TYPE; break;
            case 'Z': arg = Boolean.TYPE; break;
            case '[': dim++; break;
            case 'L':
                String className = methodDesc.substring(c + 1,
                    methodDesc.indexOf(';', c));
                arg = Class.forName(className);
                c += className.length() + 1;
                break;
            }
            if (ch != '[') {
                if (dim > 0) {
                    int[] dims = new int[dim];
                    argType[i] = Array.newInstance(arg, dims).getClass();
                    dim = 0;
                } else {
                    argType[i] = arg;
                }
                i++;
            }
            c++;
        }

        return decl 
            ? clazz.getDeclaredMethod(methodName, argType)
            : clazz.getMethod(methodName, argType);
    }

    /** Process a JVM opcode. */
    public void sawOpcode(int seen) {
	int startArg;
        switch (seen) {
	case INVOKESPECIAL:
	    startArg = 2;		// XXX Not sure why this is required
	    break;
	case INVOKEINTERFACE:
	case INVOKEINTERFACE_QUICK:
	case INVOKENONVIRTUAL_QUICK:
	case INVOKESUPER_QUICK:
	case INVOKEVIRTUAL:
	case INVOKEVIRTUAL_QUICK:
	case INVOKEVIRTUAL_QUICK_W:
	case INVOKEVIRTUALOBJECT_QUICK:
	    startArg = 1;
	    break;
	case INVOKESTATIC:
	case INVOKESTATIC_QUICK:
	    startArg = 0;
	    break;
	default:
	    stack.sawOpcode(this,seen);
	    return;
	}
	int nargs = getNArgs(getDottedSigConstantOperand());
	if (DEBUG)
	    System.out.println(getDottedSigConstantOperand() + getDottedClassConstantOperand() + "." + getNameConstantOperand() + "(" + nargs + " arguments)");
	if (showPerformanceMetrics) {
	    String niceMethodName = (new MethodAnnotation(getDottedClassConstantOperand(),
		getNameConstantOperand(), getDottedSigConstantOperand(), startArg == 0)).getFullMethod();
	    uniqueInvocations.add(niceMethodName);
	    totalInvocations++;
	}

	if (nargs == 0) {
	    stack.sawOpcode(this,seen);
	    return;
	}
	try {
	    verify("edu.umd.cs.findbugs.apiv."
		+ getDottedClassConstantOperand() + "V",
		getDottedClassConstantOperand(),
		getNameConstantOperand(),
		getDottedSigConstantOperand(),
		startArg,
		nargs);

	    verify(getDottedClassConstantOperand() + "V",
		getDottedClassConstantOperand(),
		getNameConstantOperand(),
		getDottedSigConstantOperand(),
		startArg,
		nargs);
	} catch (RuntimeException e) {
	    System.err.println("Ignoring exception " + e);
	}
    }

    /** Try to verify the method for the given class name */
    private void verify(String verifyClassName, String realClassName,
        String methodName, String methodDesc, int startArg, int nargs) {
	// Load the verifier class
	Class vc = null;
	try {
	    vc = Class.forName(verifyClassName);
	    if (DEBUG)
		System.out.println("Class found: " + verifyClassName);
	} catch (ClassNotFoundException e) {
	    if (DEBUG)
		System.out.println("Unable to load " + verifyClassName + ": " + e);
	    return;
	}

	// Load the real class
	Class rc = null;
	try {
	    rc = Class.forName(realClassName);
	    if (DEBUG)
		System.out.println("Class found: " + realClassName);
	} catch (ClassNotFoundException e) {
	    if (DEBUG)
		System.out.println("Unable to load " + realClassName + ": " + e);
	    return;
	}

	// Obtain the invocation's target and argument list
	Class argType[] = new Class[nargs + 2];
	boolean argIsConstant[] = new boolean[nargs + 1];
	Object argVal[] = new Object[nargs + 2];
	Object realTarget = null;	// The actual compile-time object
	Object verifyTarget = null;	// For calling invoke
	argVal[0] = argIsConstant;
	argType[0] = argVal[0].getClass();
	argVal[1] = null;
	int i = nargs - 1, j = 0;
        for (; i >= 0; i--, j++) {
            OpcodeStack.Item si = stack.getStackItem(i);
            argVal[j + 2] = si.getConstant();
            argIsConstant[j + 1] = (argVal[j + 2] != null || si.isNull());
        }
        // Set the invocation and verification target, if applicable
        if (startArg >= 1) {
            OpcodeStack.Item si = stack.getStackItem(nargs);
            realTarget = si.getConstant();
            argIsConstant[0] = (realTarget != null || si.isNull());
            if (realTarget == null) {
                try {
                    Constructor ct = rc.getConstructor(new Class[] {});
                    realTarget = ct.newInstance(new Object[] {});
                } catch (Exception e) {
                    if (DEBUG)
                        System.out.println("Unable to find default contructor for " + realClassName);
                }
            } else {
                if (!rc.isInstance(realTarget)) {
                    realTarget = null;
                    argIsConstant[0] = false;
                }
            }
            try {
                Constructor ct = vc.getConstructor(new Class[] {});
                verifyTarget = ct.newInstance(new Object[] {});
            } catch (Exception e) {
                if (DEBUG)
                    System.out.println("Unable to find default contructor for " + verifyClassName);
                return;
            }
            argVal[1] = realTarget;
            argType[1] = realTarget == null ? rc : realTarget.getClass();
        } else {
            argIsConstant[0] = true;
            argVal[1] = null;
            argType[1] = rc;
        }

	// Locate the verifier method
	java.lang.reflect.Method vm = null;
	if (methodName.equals("<init>") || methodName.equals("<clinit>")) {
	    String ctorName = realClassName;
	    int dot = realClassName.lastIndexOf('.');
	    if (dot != -1)
		methodName = realClassName.substring(dot + 1);
	    else
		methodName = realClassName;
	    if (DEBUG)
		System.out.println("Encountered ctor; will call " + methodName);
	}
	try {
	    if (startArg == 1)
                vm = getVMethod(vc, methodName, methodDesc, argType, 2, false);
            else
                vm = getVMethod(vc, methodName, methodDesc, argType, 2, true);
	    if (DEBUG) {
		System.out.println("Found the method " + verifyClassName + "."
                    + methodName);
	    }
	} catch (Exception e) {
	    if (DEBUG)
		System.out.println("Unable to locate the method "
                    + verifyClassName + "." + methodName + ": " + e);
	    return;
	}
	if (vm == null)
	    return;

	// Verify and report
	if (DEBUG)
	    System.out.println("Ready to verify");
        // Fix arguments and constants lists (FindBugs gets the wrong
        // arguments sometimes)
        for (i = 2; i < argType.length; i++) {
            if (!argType[i].isPrimitive()) {
                if (!argType[i].isInstance(argVal[i])) {
                    argVal[i] = null;
                    argIsConstant[i-1] = false;
                }
            } else {
                if (argType[i].getName().equals("boolean")) {
                    if (argVal[i].toString().equals("1"))
                        argVal[i] = true;
                    else if (argVal[i].toString().equals("0"))
                        argVal[i] = false;
                }
            }
        }
        if (DEBUG) {
            System.out.print("Method invocation ");
            System.out.print(vc + ": " + verifyTarget +  " . ");
            System.out.print(getNameConstantOperand() + "(");
            for (i = 0; i < argType.length; i++) {
                if (i == 0)
                    System.out.print("boolean: isConstant[]");
                else
                    System.out.print(argType[i] + ": " + argVal[i] + "("
                        + argIsConstant[i - 1] + ")");
                if (i < argType.length - 1)
                    System.out.print(", ");
            }
            System.out.println(")");
        }
	VerifyResult.Value vr = VerifyResult.Value.UNVERIFIED;
	try {
	    vr = (VerifyResult.Value)(vm.invoke(verifyTarget, argVal));
	} catch (Exception e) {
	    if (DEBUG)
		System.out.println("Invocation failed: " + e);
	}
	String errorCode = null;
	if (vr == VerifyResult.Value.ARGUMENT_ERROR)
	    errorCode = "LCAPI_ARGUMENT";
	else if (vr == VerifyResult.Value.DOMAIN_ERROR)
	    errorCode = "LCAPI_DOMAIN";
	else if (vr == VerifyResult.Value.NAME_ERROR)
	    errorCode = "LCAPI_NAME";
	else if (vr == VerifyResult.Value.SYNTAX_ERROR)
	    errorCode = "LCAPI_SYNTAX";
	if (errorCode != null) {
	    bugReporter.reportBug(new BugInstance(this, errorCode, NORMAL_PRIORITY)
		.addClassAndMethod(this).addSourceLine(this)
		.addCalledMethod(realClassName, methodName, methodDesc, startArg == 0));
	    if (showInvocationParameters)
		for (i = 1; i < argType.length; i++)
		    System.out.println("arg" + i + " = " + (argIsConstant[i - 1] ? argVal[i] : "<unknown>"));
	}
	if (showPerformanceMetrics) {
	    String niceMethodName = (new MethodAnnotation(realClassName, methodName, methodDesc, startArg == 0)).getFullMethod();
	    EnumMap<VerifyResult.Value, Integer> m = performanceMetrics.get(niceMethodName);
	    if (m == null)
		m = new EnumMap<VerifyResult.Value, Integer>(VerifyResult.Value.class);
	    Integer v = m.get(vr);
	    if (v == null)
		v = new Integer(1);
	    else
		v = v + 1;
	    m.put(vr, v);
	    performanceMetrics.put(niceMethodName, m);
	}
	if (DEBUG)
	    System.out.println("Verification result: " + vr);
    }

    public void report() {
	if (!showPerformanceMetrics)
	    return;
	System.out.println("PM S totalInvocations " + totalInvocations);
	System.out.println("PM S uniqueInvocations " + uniqueInvocations.size());
	boolean humanOutput =
	    Boolean.getBoolean("apiverify.humanOutput");
	for (String methodName : performanceMetrics.keySet()) {
	    System.out.print("PM I " + methodName);
	    EnumMap<VerifyResult.Value, Integer> m = performanceMetrics.get(methodName);
	    if (humanOutput)
		System.out.print(" " + m);
	    else
		for (VerifyResult.Value vr : VerifyResult.Value.values()) {
		    Integer i = m.get(vr);
		    System.out.print(" " + (i == null ? 0 : i));
		}
	    System.out.println();
	}
	if (showAllInvocations)
	    for (String methodName : uniqueInvocations)
		System.out.println("PM XI " + methodName);
    }
}
